import tensorflow as tf


def export_graph(checkpoint_path, output_nodes):
    """
    Export a graph stored in a checkpoint as a *.pb file.
    :param checkpoint_path: The checkpoint path which should be frozen.
    :param output_nodes: The output nodes you care about as a list of strings (their names).
    :return:
    """
    if not tf.gfile.Exists(checkpoint_path):
        raise AssertionError(
            "Export directory doesn't exists. Please specify an export "
            "directory: %s" % checkpoint_path)

    if not output_nodes:
        print("You need to supply the name of a node to --output_node_names.")
        return -1

        # We retrieve our checkpoint fullpath
    checkpoint = tf.train.get_checkpoint_state(checkpoint_path)
    input_checkpoint = checkpoint.model_checkpoint_path

    # We precise the file fullname of our freezed graph
    output_graph = checkpoint_path + "/frozen_model.pb"

    # We clear devices to allow TensorFlow to control on which device it will load operations
    clear_devices = True

    # We start a session using a temporary fresh Graph
    with tf.Session(graph=tf.Graph()) as sess:
        # We import the meta graph in the current default Graph
        saver = tf.train.import_meta_graph(input_checkpoint + '.meta', clear_devices=clear_devices)

        # We restore the weights
        saver.restore(sess, input_checkpoint)

        # We use a built-in TF helper to export variables to constants
        output_graph_def = tf.graph_util.convert_variables_to_constants(
            sess,  # The session is used to retrieve the weights
            tf.get_default_graph().as_graph_def(),  # The graph_def is used to retrieve the nodes
            output_nodes  # The output node names are used to select the useful nodes
        )

        # Finally we serialize and dump the output graph to the filesystem
        with tf.gfile.GFile(output_graph, "wb") as f:
            f.write(output_graph_def.SerializeToString())
        print("%d ops in the final graph." % len(output_graph_def.node))

    return output_graph_def


def load_graph(frozen_graph_filename, input_shape, namespace_prefix="", old_input=None):
    """
    Loads a frozen graph from a *.pb file.
    :param frozen_graph_filename: The file which graph to load.
    :param input_shape: The shape of the new input you want.
    :param namespace_prefix: A namespace for your graph to live in. This is useful when having multiple models.
    :param old_input: The name of the old input tensor you want to replace.
    :return: The graph that can now be passed to a session when creating it.
    """
    # Load graph def from protobuff and import the definition
    with tf.gfile.GFile(frozen_graph_filename, "rb") as f:
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(f.read())

    if old_input is None:
        with tf.Graph().as_default() as graph:
            tf.import_graph_def(graph_def, name=namespace_prefix)
    else:
        with tf.Graph().as_default() as graph:
            tf_new_input = tf.placeholder(tf.float32, shape=input_shape, name="input_tensor")
            tf.import_graph_def(graph_def, input_map={old_input: tf_new_input}, name=namespace_prefix)
    return graph


def get_slim_init_model_fn(checkpoint_path, checkpoint_exclude_scopes=[]):
    """
    Create an init_model function for slim models in tensorflow/models git repository.

    :param checkpoint_path: The path where the checkpoint is stored.
    :param checkpoint_exclude_scopes: Which scopes should be excluded.
    :return: The init_model(hyper_params, model, session) function for the starttf.estimators.
    """
    if checkpoint_path is None:
        return None

    variables_to_restore = []
    for var in tf.contrib.slim.get_model_variables():
        excluded = False
        for exclusion in checkpoint_exclude_scopes:
            if var.op.name.startswith(exclusion):
                excluded = True
                break
        if not excluded:
            variables_to_restore.append(var)

    if tf.gfile.IsDirectory(checkpoint_path):
        checkpoint_path = tf.train.latest_checkpoint(checkpoint_path)
    else:
        checkpoint_path = checkpoint_path

    tf.logging.info('Fine-tuning from %s' % checkpoint_path)

    init_op = tf.contrib.slim.assign_from_checkpoint_fn(
        checkpoint_path,
        variables_to_restore,
        ignore_missing_vars=True)

    def init_model(hyper_params, model, session):
        session.run(init_op)

    return init_model
